"""Spy objects.

Classes in this module are heavily inspired by the
[unittest.mock library](https://docs.python.org/3/library/unittest.mock.html).
"""
from __future__ import annotations
from dataclasses import dataclass
from inspect import isclass, iscoroutinefunction
from typing import get_type_hints, Any, Callable, Dict, Optional, Tuple, Type


@dataclass(frozen=True)
class SpyCall:
    """A dataclass representing a call to a spy.

    Attributes:
        spy_id: Identifier of the spy that made the call
        args: Arguments list of the call
        kwargs: Keyword arguments list of the call
    """

    spy_id: int
    spy_name: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]

    def __str__(self) -> str:
        """Stringify the call to something human readable.

        `SpyCall(spy_id=42, spy_name="name", args=(1,), kwargs={"foo": False})`
        would stringify as `"name(1, foo=False)"`
        """
        args_list = [repr(arg) for arg in self.args]
        kwargs_list = [f"{key}={repr(val)}" for key, val in self.kwargs.items()]

        return f"{self.spy_name}({', '.join(args_list + kwargs_list)})"


CallHandler = Callable[[SpyCall], Any]


class BaseSpy:
    """Spy object base class.

    - Pretends to be another class, if another class is given as a spec
    - Lazily constructs child spies when an attribute is accessed
    """

    def __init__(
        self,
        handle_call: CallHandler,
        spec: Optional[Any] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize a BaseSpy from a call handler and an optional spec object."""
        self._name = name or (spec.__name__ if spec is not None else "spy")
        self._spec = spec
        self._handle_call: CallHandler = handle_call
        self._spy_children: Dict[str, BaseSpy] = {}

    @property  # type: ignore[misc]
    def __class__(self) -> Any:
        """Ensure Spy can pass `instanceof` checks."""
        if isclass(self._spec):
            return self._spec

        return type(self)

    def __getattr__(self, name: str) -> Any:
        """Get a property of the spy.

        Lazily constructs child spies, basing them on type hints if available.
        """
        if name in self._spy_children:
            return self._spy_children[name]

        child_spec = None

        if isclass(self._spec):
            hints = get_type_hints(self._spec)  # type: ignore[arg-type]
            child_spec = getattr(
                self._spec,
                name,
                hints.get(name),
            )

        if isinstance(child_spec, property):
            hints = get_type_hints(child_spec.fget)
            child_spec = hints.get("return")

        spy = create_spy(
            handle_call=self._handle_call,
            spec=child_spec,
            name=f"{self._name}.{name}",
        )

        self._spy_children[name] = spy

        return spy


class Spy(BaseSpy):
    """An object that records all calls made to itself and its children."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy."""
        return self._handle_call(SpyCall(id(self), self._name, args, kwargs))


class AsyncSpy(BaseSpy):
    """An object that records all async. calls made to itself and its children."""

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy asynchronously."""
        return self._handle_call(SpyCall(id(self), self._name, args, kwargs))


def create_spy(
    handle_call: CallHandler,
    spec: Optional[Any] = None,
    is_async: bool = False,
    name: Optional[str] = None,
) -> Any:
    """Create a Spy from a spec.

    Functions and classes passed to `spec` will be inspected (and have any type
    annotations inspected) to ensure `AsyncSpy`'s are returned where necessary.
    """
    _SpyCls: Type[BaseSpy] = Spy

    if iscoroutinefunction(spec) or is_async is True:
        _SpyCls = AsyncSpy

    return _SpyCls(handle_call=handle_call, spec=spec, name=name)



"""Rehearsed test double stub."""
from collections import deque
from typing import Deque, Generic, Optional

from .spy import SpyCall
from .types import ReturnT


class Stub(Generic[ReturnT]):
    """A rehearsed test double stub that may perform an action."""

    _rehearsal: SpyCall
    _values: Deque[ReturnT]
    _error: Optional[Exception]

    def __init__(self, rehearsal: SpyCall) -> None:
        """Initialize the stub from a rehearsal call.

        Arguments:
            rehearsal: A call tuple to match against.
        """
        self._rehearsal = rehearsal
        self._values = deque()
        self._error = None

    def then_return(self, *values: ReturnT) -> None:
        """Set the stub's return value(s).

        See [stubbing](/#stubbing) for more details.

        Arguments:
            *values: Zero or more return values. Multiple values will result
                     in different return values for subsequent calls, with the
                     last value latching in once all other values have returned.
        """
        self._error = None
        self._values = deque(values)

    def then_raise(self, error: Exception) -> None:
        """Set the stub's error value.

        See [stubbing](/#stubbing) for more details.

        Arguments:
            error: The error to raise.

        Note:
            Setting a stub to raise will prevent you from writing new
            rehearsals, because they will raise. If you need to make more calls
            to `when`, you'll need to wrap your rehearsal in a `try`.
        """
        self._value = None
        self._error = error

    def _act(self) -> Optional[ReturnT]:
        """Get the stub's value or raise its exception."""
        if self._error:
            raise self._error

        if len(self._values) == 0:
            return None
        elif len(self._values) == 1:
            return self._values[0]
        else:
            # if there is more than one value saved, shift off the queue
            return self._values.popleft()



"""Decoy and stub registration module."""
from collections import deque
from weakref import finalize
from typing import Any, Deque, Dict, List, Optional

from .spy import BaseSpy, SpyCall
from .stub import Stub


class Registry:
    """Decoy and stub configuration registry.

    The registry collects weak-references to spies and spy calls created in
    order to clean up stub configurations when a spy goes out of scope.
    """

    _calls: Deque[SpyCall]

    def __init__(self) -> None:
        """Initialize a Registry."""
        self._stub_map: Dict[int, List[Stub[Any]]] = {}
        self._calls = deque()

    @property
    def last_call(self) -> Optional[SpyCall]:
        """Peek the last call in the registry's call stack."""
        if len(self._calls) > 0:
            return self._calls[-1]
        else:
            return None

    def pop_last_call(self) -> Optional[SpyCall]:
        """Pop the last call made off the registry's call stack."""
        if len(self._calls) > 0:
            return self._calls.pop()
        else:
            return None

    def get_stubs_by_spy_id(self, spy_id: int) -> List[Stub[Any]]:
        """Get a spy's stub list by identifier.

        Arguments:
            spy_id: The unique identifer of the Spy to look up.

        Returns:
            The list of stubs matching the given Spy.
        """
        return self._stub_map.get(spy_id, [])

    def get_calls_by_spy_id(self, spy_id: int) -> List[SpyCall]:
        """Get a spy's call list by identifier.

        Arguments:
            spy_id: The unique identifer of the Spy to look up.

        Returns:
            The list of calls matching the given Spy.
        """
        return [c for c in self._calls if c.spy_id == spy_id]

    def register_spy(self, spy: BaseSpy) -> int:
        """Register a spy for tracking.

        Arguments:
            spy: The spy to track.

        Returns:
            The spy's unique identifier.
        """
        spy_id = id(spy)
        finalize(spy, self._clear_spy, spy_id)
        return spy_id

    def register_call(self, call: SpyCall) -> None:
        """Register a spy call for tracking.

        Arguments:
            call: The call to track.
        """
        self._calls.append(call)

    def register_stub(self, spy_id: int, stub: Stub[Any]) -> None:
        """Register a stub for tracking.

        Arguments:
            spy_id: The unique identifer of the Spy to look up.
            stub: The stub to track.
        """
        stub_list = self.get_stubs_by_spy_id(spy_id)
        self._stub_map[spy_id] = stub_list + [stub]

    def _clear_spy(self, spy_id: int) -> None:
        """Clear all references to a given spy_id."""
        self._calls = deque([c for c in self._calls if c.spy_id != spy_id])

        if spy_id in self._stub_map:
            del self._stub_map[spy_id]



"""Decoy test double stubbing and verification library."""
from os import linesep
from typing import cast, Any, Optional, Sequence, Type

from .registry import Registry
from .spy import create_spy, SpyCall
from .stub import Stub
from .types import ClassT, FuncT, ReturnT


class Decoy:
    """Decoy test double state container."""

    _registry: Registry

    def __init__(self) -> None:
        """Initialize the state container for test doubles and stubs.

        You should initialize a new Decoy instance for every test.

        Example:
            ```python
            import pytest
            from decoy import Decoy

            @pytest.fixture
            def decoy() -> Decoy:
                return Decoy()
            ```
        """
        self._registry = Registry()

    def create_decoy(self, spec: Type[ClassT], *, is_async: bool = False) -> ClassT:
        """Create a class decoy for `spec`.

        Arguments:
            spec: A class definition that the decoy should mirror.
            is_async: Force the returned spy to be asynchronous. In most cases,
                this argument is unnecessary, since the Spy will use `spec` to
                determine if a method should be asynchronous.

        Returns:
            A spy typecast as an instance of `spec`.

        Example:
            ```python
            def test_get_something(decoy: Decoy):
                db = decoy.create_decoy(spec=Database)
                # ...
            ```

        """
        spy = create_spy(
            spec=spec, is_async=is_async, handle_call=self._handle_spy_call
        )
        self._registry.register_spy(spy)

        return cast(ClassT, spy)

    def create_decoy_func(
        self, spec: Optional[FuncT] = None, *, is_async: bool = False
    ) -> FuncT:
        """Create a function decoy for `spec`.

        Arguments:
            spec: A function that the decoy should mirror.
            is_async: Force the returned spy to be asynchronous. In most cases,
                this argument is unnecessary, since the Spy will use `spec` to
                determine if the function should be asynchronous.

        Returns:
            A spy typecast as `spec` function.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.create_decoy_func(spec=generate_unique_id)
                # ...
            ```
        """
        spy = create_spy(
            spec=spec, is_async=is_async, handle_call=self._handle_spy_call
        )
        self._registry.register_spy(spy)

        return cast(FuncT, spy)

    def when(self, _rehearsal_result: ReturnT) -> Stub[ReturnT]:
        """Create a [Stub][decoy.stub.Stub] configuration using a rehearsal call.

        See [stubbing](index.md#stubbing) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, used for typechecking.

        Returns:
            A stub to configure using `then_return` or `then_raise`.

        Example:
            ```python
            db = decoy.create_decoy(spec=Database)
            decoy.when(db.exists("some-id")).then_return(True)
            ```

        Note:
            The "rehearsal" is an actual call to the test fake. The fact that
            the call is written inside `when` is purely for typechecking and
            API sugar. Decoy will pop the last call to _any_ fake off its
            call stack, which will end up being the call inside `when`.
        """
        rehearsal = self._pop_last_rehearsal()
        stub = Stub[ReturnT](rehearsal=rehearsal)

        self._registry.register_stub(rehearsal.spy_id, stub)

        return stub

    def verify(self, _rehearsal_result: Optional[ReturnT] = None) -> None:
        """Verify a decoy was called using a rehearsal.

        See [verification](index.md#verification) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, unused.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.create_decoy_func(spec=generate_unique_id)

                # ...

                decoy.verify(gen_id("model-prefix_"))
            ```

        Note:
            The "rehearsal" is an actual call to the test fake. The fact that
            the call is written inside `verify` is purely for typechecking and
            API sugar. Decoy will pop the last call to _any_ fake off its
            call stack, which will end up being the call inside `verify`.
        """
        rehearsal = self._pop_last_rehearsal()
        all_calls = self._registry.get_calls_by_spy_id(rehearsal.spy_id)

        assert rehearsal in all_calls, self._build_verify_error(rehearsal, all_calls)

    def _pop_last_rehearsal(self) -> SpyCall:
        rehearsal = self._registry.pop_last_call()

        if rehearsal is None:
            raise ValueError("when/verify must be called with a decoy rehearsal")

        return rehearsal

    def _handle_spy_call(self, call: SpyCall) -> Any:
        self._registry.register_call(call)

        stubs = self._registry.get_stubs_by_spy_id(call.spy_id)

        for stub in reversed(stubs):
            if stub._rehearsal == call:
                return stub._act()

        return None

    def _build_verify_error(
        self, rehearsal: SpyCall, all_calls: Sequence[SpyCall]
    ) -> str:
        all_calls_len = len(all_calls)
        all_calls_plural = all_calls_len != 1
        all_calls_printout = linesep.join(
            [f"{n + 1}.\t{str(all_calls[n])}" for n in range(all_calls_len)]
        )

        return linesep.join(
            [
                "Expected call:",
                f"\t{str(rehearsal)}",
                f"Found {all_calls_len} call{'s' if all_calls_plural else ''}:",
                all_calls_printout,
            ]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['decoy']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'decoy',
    'version': '1.1.0',
    'description': 'Opinionated, typed stubbing and verification library for Python',
    'long_description': '# Decoy\n\n[![ci badge][]][ci]\n[![pypi version badge][]][pypi]\n[![license badge][]][license]\n\n[ci]: https://github.com/mcous/decoy/actions\n[ci badge]: https://flat.badgen.net/github/checks/mcous/decoy/main\n[pypi]: https://pypi.org/project/decoy/\n[pypi version badge]: https://flat.badgen.net/pypi/v/decoy\n[license]: https://github.com/mcous/decoy/blob/main/LICENSE\n[license badge]: https://flat.badgen.net/github/license/mcous/decoy\n\n> Opinionated, typed stubbing and verification library for Python\n\n<https://mike.cousins.io/decoy/>\n\nThe Decoy library allows you to create, stub, and verify test double objects for your Python unit tests, so your tests are:\n\n-   Less prone to insufficient tests due to unconditional stubbing\n-   Covered by typechecking\n-   Easier to fit into the Arrange-Act-Assert pattern\n\nThe Decoy API is heavily inspired by / stolen from the excellent [testdouble.js][] and [Mockito][] projects.\n\n[testdouble.js]: https://github.com/testdouble/testdouble.js\n[mockito]: https://site.mockito.org/\n\n## Install\n\n```bash\n# pip\npip install decoy\n\n# poetry\npoetry add --dev decoy\n```\n\n## Setup\n\nYou\'ll want to create a test fixture to reset Decoy state between each test run. In [pytest][], you can do this by using a fixture to create a new Decoy instance for every test.\n\nThe examples below assume the following global test fixture:\n\n```python\nimport pytest\nfrom decoy import Decoy\n\n@pytest.fixture\ndef decoy() -> Decoy:\n    return Decoy()\n```\n\nWhy is this important? The `Decoy` container tracks every fake that is created during a test so that you can define assertions using fully-typed rehearsals of your test double. It\'s important to wipe this slate clean for every test so you don\'t leak memory or have any state preservation between tests.\n\n[pytest]: https://docs.pytest.org/\n\n### Mypy Setup\n\nDecoy\'s rehearsal syntax can be a bit confusing to [mypy][] if the mock in question is supposed to return `None`. Normally, [mypy will complain][] if you try to use a `None`-returning expression as a value, because this is almost always a mistake.\n\nIn Decoy, however, it\'s an intentional part of the API and _not_ a mistake. To suppress these errors, Decoy provides a mypy plugin that you should add to your configuration file:\n\n```ini\n# mypi.ini\n\n# ...\nplugins = decoy.mypy\n# ...\n```\n\n[mypy]: https://mypy.readthedocs.io/\n[mypy will complain]: https://mypy.readthedocs.io/en/stable/error_code_list.html#check-that-called-function-returns-a-value-func-returns-value\n\n## Usage\n\n### Stubbing\n\nA stub is a an object used in a test that is pre-configured to return a result or raise an error if called according to a specification. In Decoy, you specify a stub\'s call expectations with a "rehearsal", which is simply a call to the stub inside of a `decoy.when` wrapper.\n\nBy pre-configuring the stub with specific rehearsals, you get the following benefits:\n\n-   Your test double will only return your mock value **if it is called correctly**\n-   You avoid separate "set up mock return value" and "assert mock called correctly" steps\n-   If you annotate your test double with an actual type, the rehearsal will fail typechecking if called incorrectly\n\n```python\nimport pytest\nfrom typing import cast, Optional\nfrom decoy import Decoy\n\nfrom .database import Database, Model\n\ndef get_item(uid: str, db: Database) -> Optional[Model]:\n  return db.get_by_id(uid)\n\ndef test_get_item(decoy: Decoy):\n    mock_item = cast(Model, { "foo": "bar" })\n    mock_db = decoy.create_decoy(spec=Database)\n\n    # arrange stub using rehearsals\n    decoy.when(mock_db.get_by_id("some-id")).then_return(mock_item)\n\n    # call code under test\n    some_result = get_item("some-id")\n    other_result = get_item("other-id")\n\n    # assert code result\n    assert some_result == mock_item\n    assert other_result is None\n```\n\n### Verifying interactions\n\nIf you\'re coming from `unittest.mock`, you\'re probably used to calling your code under test and _then_ verifying that your dependency was called correctly. Decoy provides similar call verification using the same "rehearsal" mechanism that the stubbing API uses.\n\n```python\nimport pytest\nfrom typing import cast, Optional\nfrom decoy import Decoy, verify\n\nfrom .logger import Logger\n\ndef log_warning(msg: str, logger: Logger) -> None:\n    logger.warn(msg)\n\ndef test_log_warning(decoy: Decoy):\n    logger = decoy.create_decoy(spec=Logger)\n\n    # call code under test\n    some_result = log_warning("oh no!", logger)\n\n    # verify double called correctly with a rehearsal\n    decoy.verify(logger.warn("oh no!"))\n```\n\nAsserting that calls happened after the fact can be useful, but **should only be used if the dependency is being called solely for its side-effect(s)**. Verification of interactions in this manner should be considered a last resort, because:\n\n-   If you\'re calling a dependency to get data, then you can more precisely describe that relationship using [stubbing](#stubbing)\n-   Side-effects are harder to understand and maintain than pure functions, so in general you should try to side-effect sparingly\n\nStubbing and verification of a decoy are **mutually exclusive** within a test. If you find yourself wanting to both stub and verify the same decoy, then one or more of these is true:\n\n-   The assertions are redundant\n-   The dependency is doing too much based on its input (e.g. side-effecting _and_ calculating complex data) and should be refactored\n\n### Usage with async/await\n\nDecoy supports async/await out of the box! Pass your async function or class with async methods to `spec` in `decoy.create_decoy_func` or `decoy.create_decoy`, respectively, and Decoy will figure out the rest.\n\nWhen writing rehearsals on async functions and methods, remember to include the `await` with your rehearsal call:\n\n```py\ndecoy.when(await mock_db.get_by_id("some-id")).then_return(mock_item)\n```\n\n### Matchers\n\nSometimes, when you\'re stubbing or verifying calls (or really when you\'re doing any sort of equality assertion in a test), you need to loosen a given assertion. For example, you may want to assert that a dependency is called with a string, but you don\'t care about the full contents of that string.\n\nDecoy includes a set of matchers, which are simply Python classes with `__eq__` methods defined, that you can use in rehearsals and/or assertions.\n\n```python\nimport pytest\nfrom typing import cast, Optional\nfrom decoy import Decoy, matchers\n\nfrom .logger import Logger\n\ndef log_warning(msg: str, logger: Logger) -> None:\n    logger.warn(msg)\n\ndef test_log_warning(decoy: Decoy):\n    logger = decoy.create_decoy(spec=Logger)\n\n    # call code under test\n    some_result = log_warning(\n        "Oh no, something went wrong with request ID abc123efg456",\n        logger=logger\n    )\n\n    # verify double called correctly\n    decoy.verify(\n        logger.warn(matchers.StringMatching("request ID abc123efg456"))\n    )\n```\n',
    'author': 'Mike Cousins',
    'author_email': 'mike@cousins.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://mike.cousins.io/decoy/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
